package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.Conjunction
import ap.util.{Debug, Logic}

object AndTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(left : ProofTree, right : ProofTree,
            order : TermOrder,
            simplifier : ConstraintSimplifier) : ProofTree = {
    val heightDiff = heightOf(right) - heightOf(left)
    if (heightDiff > 1) {
      // then right is an <code>AndTree</code> that is very deep
      val AndTree(rightLeft, rightRight) = right
      if (heightOf(rightLeft) < heightOf(rightRight)) {
        // left-rotate
        AndTree(AndTree(left, rightLeft, order, simplifier),
                rightRight,
                order, simplifier)
      } else {
        // right-left-rotate
        val AndTree(rightLeftLeft, rightLeftRight) = rightLeft
        AndTree(AndTree(left, rightLeftLeft, order, simplifier),
                AndTree(rightLeftRight, rightRight, order, simplifier),
                order, simplifier)
      }
    } else if (heightDiff < -1) {
      // then left is an <code>AndTree</code> that is very deep
      val AndTree(leftLeft, leftRight) = left
      if (heightOf(leftRight) < heightOf(leftLeft)) {
        // right-rotate
        AndTree(leftLeft,
                AndTree(leftRight, right, order, simplifier),
                order, simplifier)
      } else {
        // left-right-rotate
        val AndTree(leftRightLeft, leftRightRight) = leftRight
        AndTree(AndTree(leftLeft, leftRightLeft, order, simplifier),
                AndTree(leftRightRight, right, order, simplifier),
                order, simplifier)
      }
    } else {
      new AndTree(left, right, order, simplifier)
    }
  }

  def apply(subtrees : Seq[ProofTree], order : TermOrder,
            simplifier : ConstraintSimplifier) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, !subtrees.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    subtrees match {
      case Seq(subtree) => subtree
      case Seq(left, right) => AndTree(left, right, order, simplifier)
      case _ => {
        // divide the subtrees in two parts and recurse
        val mid = (subtrees.size + 1) / 2
        AndTree(AndTree(subtrees.slice(0, mid), order, simplifier),
                AndTree(subtrees.slice(mid, subtrees.size), order, simplifier),
                order, simplifier)
      }
    }
  }
    
  def unapply(t : ProofTree) : Option[(ProofTree, ProofTree)] = t match {
    case t : AndTree => Some((t.left, t.right))
    case _ => None
  }
  
  private def heightOf(t : ProofTree) : Int = t match {
    case t : AndTree => t.height
    case _ => 0
  }
}

class AndTree private (val left : ProofTree, val right : ProofTree,
                       protected val order : TermOrder,
                       simplifier : ConstraintSimplifier) extends ProofTree {

  import AndTree.heightOf
  
  //////////////////////////////////////////////////////////////////////////////
  // Branching points in proof trees are represented as balanced binary trees
  Debug.assertCtor(AndTree.AC, (heightOf(left) - heightOf(right)).abs <= 1)
  //////////////////////////////////////////////////////////////////////////////
  
  lazy val height : Int = 1 + (heightOf(left) max heightOf(right))

  lazy val subtrees : Seq[ProofTree] = Array(left, right)
    
  protected[proof] lazy val rawClosingConstraint : Conjunction =
    Conjunction.conj(for (t <- subtrees.elements) yield t.closingConstraint,
                     order)

  lazy val closingConstraint : Conjunction = simplifier(rawClosingConstraint, order)

  lazy val stepPossible : Boolean =
    Logic.exists(for (subtree <- subtrees.elements) yield subtree.stepPossible)

  lazy val stepMeaningful : Boolean =
    (left.stepMeaningful, right.stepMeaningful) match {
    case (true, true) => true
    case (true, false) => !right.closingConstraint.isFalse
    case (false, true) => !left.closingConstraint.isFalse
    case (false, false) => false
    }

  def update(newSubtrees : Seq[ProofTree]) : ProofTree =
    AndTree(newSubtrees, order, simplifier)

  def update(newLeft : ProofTree, newRight : ProofTree) : ProofTree =
    if (left == newLeft && right == newRight)
      this
    else
      AndTree(newLeft, newRight, order, simplifier)
    
  //////////////////////////////////////////////////////////////////////////////

  private def indent(str : String, prefix : String) =
    prefix + str.replaceAll("\n", "\n" + prefix)
    
  override def toString : String = {
    "AndTree (" + closingConstraint + ") {\n" +
      (for (t <- subtrees) yield indent("" + t, "  ")).reduceLeft(
         (s1 : String, s2 : String) => s1 + "\n/\\\n" + s2) +
    "\n}"
  }

}
