package ap.proof;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.proof.goal.Goal
import ap.util.{Debug, Seqs, LRUCache}

object ConstraintSimplifier {
  
  val AC = Debug.AC_CONSTRAINT_SIMPLIFIER  
  
  val NO_SIMPLIFIER : ConstraintSimplifier = new ConstraintSimplifier {
    def apply(f : Conjunction, order : TermOrder) : Conjunction = f
    def canHandle(f : Conjunction) : Boolean = true
  }
  
  val ALL_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(true, false)

  val ALL_SIMPLIFIER_TRACE : ConstraintSimplifier =
    new SimpleSimplifier(true, true)

  val QUANTIFICATION_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(false, false)

}

abstract class ConstraintSimplifier {
  /**
   * Simplify the formula <code>f</code>
   */
  def apply(f : Conjunction, order : TermOrder) : Conjunction
  
  def canHandle(f : Conjunction) : Boolean
}

class SimpleSimplifier(groundSimplifications : Boolean, output : Boolean)
      extends ConstraintSimplifier {
  
  private lazy val prover =
    new ExhaustiveProver(ConstraintSimplifier.NO_SIMPLIFIER)

  private def print(str : => String) : unit = (if (output) Predef.print(str))
  private def println(str : => String) : unit = (if (output) Predef.println(str))
  
  private val EMPTY : Set[Quantifier]  = Set()
  private val ALL : Set[Quantifier]    = Set(Quantifier.ALL)
  private val EX : Set[Quantifier]     = Set(Quantifier.EX)
  private val ALL_EX : Set[Quantifier] = Set(Quantifier.EX, Quantifier.ALL)
    
  def canHandle(f : Conjunction) : Boolean = collectQuantifiers(f) != ALL_EX

  private val cache = new LRUCache[Conjunction, Conjunction] (1000)
  
  def apply(f : Conjunction, order : TermOrder) : Conjunction = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConstraintSimplifier.AC, order isSortingOf f)
    ////////////////////////////////////////////////////////////////////////////
    cache(f) {
      collectQuantifiers(f) match {
      case ALL => simplify(f, order)
      case EX => negSimplify(f, order)
      case EMPTY if (groundSimplifications) =>
        if (f.size > 1)
          negSimplify(f, order)
        else
          f
      case EMPTY => f
      }
    }
  }
  
  private def simplify(f : Conjunction, order : TermOrder) : Conjunction = {
    println("Simplify: " + f + " (positive)")
    val res = prover(f, order).closingConstraint
    println("          " + res)
    res
  }

  private def negSimplify(f : Conjunction, order : TermOrder) : Conjunction = {
    println("Simplify: " + f + " (negative)")
    val res = prover(f.negate, order).closingConstraint.negate    
    println("          " + res)
    res
  }
  
  private def collectQuantifiers(f : Formula) : Set[Quantifier] =
    Conjunction.collectQuantifiers(f, (conj) =>
           Set() ++ conj.quans.drop(1) ++
           // we only look for proper divisibility (not divisibility by one ...)
           (if (((conj.arithConj.positiveEqs.elements ++
                  conj.arithConj.negativeEqs.elements).next()(0) _1).isOne)
              Set(conj.quans(0))
            else
              Set()))
  
}
