package ap.proof;

import ap.terfor.{TermOrder, ConstantTerm, VariableTerm, Formula}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.util.{PlainRange, Debug, Seqs}

object QuantifiedTree {
  
  private val AC = Debug.AC_PROOF_TREE
  
  def apply(subtree : ProofTree,
            quan : Quantifier,
            quantifiedConstants : Set[ConstantTerm],
            order : TermOrder,
            subtreeOrder : TermOrder,
            simplifier : ConstraintSimplifier) : ProofTree = subtree match {
    case subtree : QuantifiedTree if (subtree.quan == quan) =>
      QuantifiedTree(subtree.subtree,
                     quan,
                     quantifiedConstants ++ subtree.quantifiedConstants,
                     order, subtree.subtreeOrder,
                     simplifier)
    case subtree : AndTree if (quan == Quantifier.ALL) =>
      AndTree(for (t <- subtree.subtrees)
              yield QuantifiedTree(t, quan, quantifiedConstants, order,
                                   subtreeOrder, simplifier),
              order, simplifier)
    case subtree : WeakenTree if (quan == Quantifier.EX) =>
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                order, subtreeOrder, simplifier),
                 quantifySimplify(subtree.disjunct, quan, quantifiedConstants,
                                  order, subtreeOrder, simplifier),
                 order)
    case subtree : WeakenTree if (Seqs.disjoint(quantifiedConstants,
                                                subtree.disjunct.constants)) =>
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                order, subtreeOrder, simplifier),
                 order sort subtree.disjunct,
                 order)
    case _ => 
      new QuantifiedTree(subtree, quan, quantifiedConstants, order, subtreeOrder,
                         simplifier)
  }
 
  def unapply(tree : ProofTree) : Option[(Quantifier, ProofTree)] = tree match {
    case tree : QuantifiedTree => Some(tree.quan, tree.subtree)
    case _ => None
  }
  
  private def quantify(f : Conjunction,
                       quan : Quantifier, quantifiedConstants : Set[ConstantTerm],
                       outputOrder : TermOrder, inputOrder : TermOrder) : Conjunction =
    Conjunction.quantify(quan, quantifiedConstants,
                         f, inputOrder) sortBy outputOrder
  
  private def quantifySimplify(f : Conjunction,
                               quan : Quantifier, quantifiedConstants : Set[ConstantTerm],
                               outputOrder : TermOrder, inputOrder : TermOrder,
                               simplifier : ConstraintSimplifier) : Conjunction =  
    simplifier(quantify(f, quan, quantifiedConstants, outputOrder, inputOrder),
               outputOrder)
  
}

/**
 * <code>ProofTreeOneChild</code> that quantifies a set of constants in
 * the closing constraint of its <code>subtree</code>
 */
class QuantifiedTree private (val subtree : ProofTree,
                              val quan : Quantifier,
                              val quantifiedConstants : Set[ConstantTerm],
                              protected val order : TermOrder,
                              protected val subtreeOrder : TermOrder,
                              simplifier : ConstraintSimplifier)
      extends ProofTreeOneChild {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifiedTree.AC,
                   !quantifiedConstants.isEmpty &&
                   (subtree match {
                     case subtree : QuantifiedTree => subtree.quan != this.quan
                     case subtree : AndTree => this.quan != Quantifier.ALL
                     case subtree : WeakenTree if (this.quan == Quantifier.EX) =>
                       false
                     case subtree : WeakenTree if (this.quan == Quantifier.ALL) =>
                       !(quantifiedConstants intersect subtree.disjunct.constants).isEmpty
                     case _ => true
                    }))
  //////////////////////////////////////////////////////////////////////////////

  protected[proof] lazy val rawClosingConstraint : Conjunction = {
    val firstTry =
      QuantifiedTree.quantify(subtree.rawClosingConstraint, quan, quantifiedConstants,
                              order, subtreeOrder)
    if (simplifier canHandle firstTry)
      firstTry
    else
      QuantifiedTree.quantify(subtree.closingConstraint, quan, quantifiedConstants,
                              order, subtreeOrder)
  }

  lazy val closingConstraint : Conjunction =
    simplifier(rawClosingConstraint, order)
  
  lazy val stepMeaningful : Boolean = subtree.stepMeaningful

  def update(newSubtree : ProofTree) : ProofTree =
    if (subtree == newSubtree)
      this
    else
      QuantifiedTree(newSubtree, quan, quantifiedConstants, order, subtreeOrder,
                     simplifier)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    "QuantifiedTree (" + quan + ", " + quantifiedConstants + ") (" +
      closingConstraint + ") {\n" + subtree + "\n}"
  }

}
