package ap.proof.goal;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.{Formula, ConstantTerm, TerFor}
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.substitutions.VariableSubst
import ap.terfor.preds.PredConj
import ap.util.{Debug, PlainRange, Seqs, FilterIt}
import ap.parameters.Param

import scala.collection.mutable.ArrayBuffer

object BetaFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    f.size > 1 && f.quans.isEmpty

}


class BetaFormulaTask(_formula : Conjunction, _age : Int,
                      symbolWeights : SymbolWeights)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(BetaFormulaTask.AC,
                   !formula.isTrue && !formula.isFalse &&
                   !formula.isLiteral && !formula.isNegatedConjunction)
  //////////////////////////////////////////////////////////////////////////////
      
  val priority : Int =
    -400 - (symbolWeights maxWeight formula) / 30 +
    formula.arithConj.size * 50 +
    formula.predConj.size * 10 +
    formula.negatedConjs.size * 50

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (!splittingNecessary(goal)) {
      // we just add the formula as a clause to the goal
      val newClauses = NegatedConjunctions(goal.clauses.elements ++
                                           Iterator.single(formula),
                                           goal.order)
      ptf.updateGoal(newClauses, goal)
    } else {
      if (formula.arithConj.isTrue && formula.predConj.isTrue) {
        // split to handle the complex conjuncts
        applyBeta(goal, ptf)
      } else {
        // cut over one of the literals
        applyCut(goal, ptf)
      }
    }

  /**
   * Determine whether this formula requires real splitting, or whether it can
   * be passed to the constraint unchanged
   */
  def splittingNecessary(goal : Goal) : Boolean =
    !formula.predicates.isEmpty ||
    !Conjunction.collectQuantifiers(formula).isEmpty ||
    !Seqs.disjoint(goal.eliminatedConstants, formula.constants)
      
  private def applyBeta(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, formula.negatedConjs.size > 1)
    ////////////////////////////////////////////////////////////////////////////
      
    val subtrees = for (negConj <- formula.negatedConjs)
                   yield ptf.updateGoal(goal.formulaTasks(negConj.negate), goal)
    ptf.and(subtrees, goal.order)
  }

  private def applyCut(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val order = goal.order    
    val (cutLiteral, rem) = selectCutLiteral(formula, goal)

    val firstTree = {
      // assume that the cut-literal is true
      val remTasks = goal.formulaTasks(rem)
      val allTasks =
        if (introduceLemma(cutLiteral, goal))
          // make a real cut
          remTasks ++ goal.formulaTasks(cutLiteral.negate)
        else
          remTasks
          
      ptf.updateGoal(allTasks, goal)
    }
    
    val secondTree = {
      // assume that the cut-literal is false
      ptf.updateGoal(goal.formulaTasks(cutLiteral), goal)
    }
    
    ptf.and(Array(secondTree, firstTree), order)    
  }
  
  private def introduceLemma(cutLiteral : Conjunction, goal : Goal) : Boolean =
    if (cutLiteral.predConj.isTrue)
      // the cut-literal is an arithmetic literal
      (if (cutLiteral.arithConj.negativeEqs.isEmpty)
         !Seqs.disjoint(goal.eliminatedConstants, cutLiteral.constants)
       else
         cutLiteral.constants subsetOf goal.eliminatedConstants)
    else
      // the cut-literal is a predicate literal
      cutLiteral.constants subsetOf goal.eliminatedConstants
  
  //////////////////////////////////////////////////////////////////////////////
      
  /**
   * Given a conjunction of formulas, select one of the formulas
   * as literal to cut over. The result is the pair of the selected formula and
   * the remaining <code>Conjunction</code>
   */
  private def selectCutLiteral(conj : Conjunction, goal : Goal)
                                              : (Conjunction, Conjunction) = {
    val weights = Param.SYMBOL_WEIGHTS(goal.settings)
    if (conj.predConj.isTrue) {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(BetaFormulaTask.AC, !conj.arithConj.isTrue)
      //////////////////////////////////////////////////////////////////////////
      val (sel, rem) =
        selectCutLiteral(conj.arithConj, goal.eliminatedConstants, weights)
      (Conjunction.conj(sel, conj.order), conj.updateArithConj(rem, conj.order))      
    } else {
      val (sel, rem) = selectHeaviestLiteral(conj.predConj, (p) => weights maxWeight p)
      (Conjunction.conj(sel, goal.order), conj.updatePredConj(rem, conj.order))
    }
  }

  //////////////////////////////////////////////////////////////////////////////
   
  private def selectHeaviestLiteral(conj : PredConj, weighter : (PredConj) => Int)
                                                 : (PredConj, PredConj) = {
    val (bestLit, remainingLits) = selectHeaviestLiteral(conj.elements, weighter)    
    (bestLit, PredConj.conj(remainingLits, conj.order))
  }

  private def selectHeaviestLiteral[A <: TerFor]
              (lits : Iterator[A], weighter : (A) => Int) : (A, ArrayBuffer[A]) = {
    val remainingLits = new ArrayBuffer[A]
    var bestLit : A = lits.next  // lits has to be non-empty
    var bestVal : Int = weighter(bestLit)
    
    for (c <- lits) {
      val newVal = weighter(c)
      if (newVal > bestVal) {
        remainingLits += bestLit
        bestLit = c
        bestVal = newVal
      } else {
        remainingLits += c
      }
    }
    
    (bestLit, remainingLits)
  }
  
  //////////////////////////////////////////////////////////////////////////////

  private def selectCutLiteralXX(ac : ArithConj,
                               eliminatedConstants : Set[ConstantTerm],
                               weights : SymbolWeights)
                                                : (ArithConj, ArithConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !ac.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val (bestLit, remainingLits) =
      selectHeaviestLiteral(ac.elements, (p:ArithConj) => weights maxWeight p)
    (bestLit, ArithConj.conj(remainingLits, ac.order))    
  }

  private def selectCutLiteral(ac : ArithConj,
                               eliminatedConstants : Set[ConstantTerm],
                               weights : SymbolWeights)
                                         : (ArithConj, ArithConj) = {
    def selectPositiveEq = {
      val (sel, rem) = selectCutLiteral(ac.positiveEqs, weights)
      (ArithConj.conj(sel, ac.order), ac.updatePositiveEqs(rem, ac.order))      
    }
    def selectNegativeEq = {
      val (sel, rem) = selectCutLiteral(ac.negativeEqs, weights)
      (ArithConj.conj(sel, ac.order), ac.updateNegativeEqs(rem, ac.order))      
    }
    def selectInEq = {
      val (sel, rem) = selectCutLiteral(ac.inEqs, weights)
      (ArithConj.conj(sel, ac.order), ac.updateInEqs(rem, ac.order))      
    }
    
    // try to select a cut-literal that contains eliminated constants
    (ac.positiveEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.positiveEqs.constants),
     ac.negativeEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.negativeEqs.constants),
     ac.inEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.inEqs.constants)) match {
      
      case (false, false, _, _, _, _) => selectPositiveEq
      case (_, _, _, _, false, false) => selectInEq
      case (_, _, false, false, _, _) => selectNegativeEq
      case (false, _, _, _, _, _)     => selectPositiveEq
      case (_, _, _, _, false, _)     => selectInEq
      case (_, _, false, _, _, _)     => selectNegativeEq
      
    }
  }

  private def selectCutLiteral(eqs : EquationConj, weights : SymbolWeights)
                                           : (EquationConj, EquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val (bestLit, remainingLits) =
      selectHeaviestLiteral(eqs.elements, lcWeighter(weights))
    (EquationConj(bestLit, eqs.order), eqs.updateEqs(remainingLits, eqs.order))
  }

  private def selectCutLiteral(eqs : NegEquationConj, weights : SymbolWeights)
                                      : (NegEquationConj, NegEquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val (bestLit, remainingLits) =
      selectHeaviestLiteral(eqs.elements, lcWeighter(weights))
    (NegEquationConj(bestLit, eqs.order),
     eqs.updateEqs(remainingLits, eqs.order))
  }
  
  private def selectCutLiteral(inEqs : InEqConj, weights : SymbolWeights)
                                                   : (InEqConj, InEqConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !inEqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    val (bestLit, remainingLits) =
      selectHeaviestLiteral(inEqs.elements, lcWeighter(weights))
    (InEqConj(bestLit, inEqs.order), InEqConj(remainingLits, inEqs.order))
  }

  //////////////////////////////////////////////////////////////////////////////

  private def size(eqs : Iterable[LinearCombination]) : Int =
    (0 /: (for (lc <- eqs.elements) yield lc.size)) ((s:Int, n:Int) => s+n)

  private def size(ac : ArithConj) : Int =
    size(ac.positiveEqs) + size(ac.negativeEqs) + size(ac.inEqs)
    
  private def lcWeighter(weights : SymbolWeights) =
    (lc : LinearCombination) => (weights maxWeight lc) - lc.size * 100
    
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction, goal : Goal) : FormulaTask =
    new BetaFormulaTask(f, age, Param.SYMBOL_WEIGHTS(goal.settings))

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    BetaFormulaTask.isCoveredFormula(f)

  protected val name : String = "BetaFor"

}
